/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2019 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#ifndef LIBFAWK_DATA_H
#define LIBFAWK_DATA_H

#define FAWK_API_VER 1

#include <stddef.h>
#include <libfawk/config.h>

typedef struct fawk_cell_s fawk_cell_t;
typedef struct fawk_ctx_s fawk_ctx_t;

#include <genht/htsp.h>
#include <genht/htpp.h>

typedef enum {
	FAWK_NIL,
	FAWK_NUM,
	FAWK_STR,
	FAWK_STRNUM,
	FAWK_ARRAY,
	FAWK_FUNC,
	FAWK_SYMREF,   /* symtab reference: reference by name */

	/* marker for returning from a C call */
	FAWK_CCALL_RET,

	/* classes for symtab: FAWK_SCALAR, FAWK_ARRAY, FAWK_FUNC */
	FAWK_SCALAR = FAWK_NIL
} fawk_celltype_t;

typedef struct {
	fawk_num_t num; /* for string this is the string converted; */
	fawk_refco_t refco;
	size_t used, alloced; /* not including the \0 (allocation is +1 in size) */
	char str[1]; /* extends beyond the srtuct */
} fawk_str_t;

typedef struct {
	fawk_celltype_t type; /* FAWK_NUM, FAWK_STRNUM or FAWK_STR */
	union {
		fawk_num_t num;
		fawk_str_t *str;
	} data;
} fawk_arridx_t;

typedef struct {
	fawk_refco_t refco;
	long uid;
	htpp_t hash;  /* fawk_arridx_t -> value-cell hash */
	unsigned destroying:1;
} fawk_arr_t;

/* Script callable functions implemnented in C shall have this prototype: */
typedef void (*fawk_cfunc_t)(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval);

/* A function implemented in script */
typedef struct {
	const char *name; /* points to a cell name */
	fawk_cfunc_t cfunc;
	size_t ip;
	int numargs, numfixedargs;
} fawk_func_t;

/* A named symbol (variable) */
typedef struct {
	union {
		fawk_cell_t *global; /* points to a symtab entry */
		int local;           /* relative sp to a local var; fp+local is the absolute address on the stack */
	} ref;
	char is_local; /* selects addr.sp over addr.global */
	size_t idx_len; /* length of idx; 0 for scalar; has to be size_t because this type is used for for(in) index lists */
	fawk_arridx_t *idx;
} fawk_symref_t;

struct fawk_cell_s {
	char *name;  /* points to the hash key if global */
	fawk_celltype_t type;
	union {
		fawk_num_t num;  /* pure number */
		fawk_str_t *str; /* string or strnum */
		fawk_arr_t *arr; /* an array of cells */
		fawk_symref_t symref;
		fawk_func_t func;
	} data;
};

typedef enum {
	FAWKI_MAKE_SYMREF, /* take entries from the stack and replace with a symref cell */
	FAWKI_PUSH_NUM,
	FAWKI_PUSH_STR,
	FAWKI_PUSH_SYMVAL, /* take a symref from the top of the stack and replace it with the sym value; same as topvar but removes the var ref from stack */
	FAWKI_PUSH_TOPVAR, /* deref the var on stack-top and push its value without removing the symref; same as symval but doesn't remove the var ref from stack */
	FAWKI_PUSH_REL,    /* take a previous stack element, relatively addressed from current sp (-1 means top) and copy-push it on top */
	FAWKI_PUSH_NIL,
	FAWKI_POP,
	FAWKI_POPJZ,
	FAWKI_POPJNZ,
	FAWKI_NEG,
	FAWKI_NOT,
	FAWKI_EQ,
	FAWKI_NEQ,
	FAWKI_LTEQ,
	FAWKI_GTEQ,
	FAWKI_LT,
	FAWKI_GT,
	FAWKI_IN,
	FAWKI_ADD,
	FAWKI_SUB,
	FAWKI_MUL,
	FAWKI_DIV,
	FAWKI_MOD,
	FAWKI_CONCAT,
	FAWKI_FORIN_FIRST,
	FAWKI_FORIN_NEXT,
	FAWKI_INCDEC,
	FAWKI_SET,
	FAWKI_CALL,
	FAWKI_RET,
	FAWKI_JMP,
	FAWKI_ABORT
} fawk_ins_t;

/* the bytecode is built of fawk_code_t entries */
typedef struct {
	enum {
		FAWKC_INS,
		FAWKC_SYMREF,
		FAWKC_NUM,
		FAWKC_STR,
		FAWKC_CSTR  /* constant string - do not free */
	} type;
	union {
		fawk_ins_t ins;
		fawk_symref_t *symref;
		fawk_num_t num;
		fawk_str_t *str;  /* string literal allocated during compilation */
		const char *cstr; /* constant string literal stored during compilation */
	} data;
	unsigned int line;
} fawk_code_t;

/* Number of cells on each stack page allocated */
#define FAWK_STACK_PAGE_SIZE 256

/* Maximum depth of #includes */
#define FAWK_MAX_INCLUDE_STACK 16

/* Parsers keep track of source files using: */
typedef struct fawk_src_s {
	char *fn;        /* upon include, filled in by libfawk; fawk_free()'d on EOF */
	long line, col;  /* line and column number within the file being parsed, both counting from 0 */
	void *user_data; /* filed in by the application */
} fawk_src_t;

struct fawk_ctx_s {
	htsp_t symtab;     /* global symbols: name->cell */

	struct {
		int (*get_char)(fawk_ctx_t *ctx, fawk_src_t *src); /* caller defined */
		int (*include)(fawk_ctx_t *ctx, fawk_src_t *src, int opening, fawk_src_t *from); /* optional: caller defined for #include support; if opening is zero, close the file; if opening is non-zero, 'from' is the current file that is executing the include */
		fawk_src_t *isp; /* include stack pointer: current file being loaded */
		fawk_src_t include_stack[FAWK_MAX_INCLUDE_STACK];
		int in_textblk, pushback; /* text block escape char and one character can be pushed back */
		char *buff, *curr_func;
		size_t used, alloced;
		unsigned textblk_state, in_eof:1;        /* in_eof: got EOF on the current level of include already */
	} parser;

	struct {
		int alloced, used;  /* number of pages allocated or in use */
		int avail;          /* available items on the current page */
		fawk_cell_t **page; /* pointer to each page */
	} stack;

	struct {
		size_t used, alloced;
		fawk_code_t *code;
	} code;

	struct {
		int numargs, numfixedargs, numidx, funcdef_offs; /* number of function arguments and array indices during parsing; BASIC: stack offset where local variables start */
		unsigned int line;
		htpp_t *labels, *lablink; /* for BASIC: label->ip, ip->label backref */
	} compiler;

	size_t errbuff_alloced;
	char *errbuff;

	/* execution */
	size_t ip;          /* instruction pointer: next instruction to process */
	size_t sp;          /* stack pointer: next (yet unused) item to write */
	size_t fp;
	long arr_uid;

	struct {
		unsigned trace:1;
		unsigned error:1;
	} exec;

	void *user_data;    /* filled in and used by the host app */
};

typedef enum {
	FAWK_ER_FIN,        /* success: function call finished; caller MUST pop the result from the stack */
	FAWK_ER_STEPS,      /* ran out of steps; more execute() is needed */
	FAWK_ER_ERROR       /* runtime error: the stack is in invalid state, reset required */
} fawk_execret_t;


/* invalid stack address */
#define FAWK_STACK_INVALID (-1)

/* invalid code address */
#define FAWK_CODE_INVALID (-1)

/* Initialize an empty script context ctx (allocated by the caller) */
FAWK_API void fawk_init(fawk_ctx_t *ctx);

/* Free all resources (stack, VM, symbols) and leave ctx empty; ctx
   itself is not free'd as it's allocated by the caller */
FAWK_API void fawk_uninit(fawk_ctx_t *ctx);

/* portable version of the standard strdup() function (whcih is not C89),
   using context-specific memory allocation */
FAWK_API char *fawk_strdup(fawk_ctx_t *ctx, const char *s);

/* Debug: print a cell to stdout; if verbose is 1, also print metadata */
FAWK_API void fawk_dump_cell(fawk_cell_t *cell, int verbose);

#endif /* LIBFAWK_DATA_H */
