/*
libfawk - A function-only AWK dialect (pascal compiler)

Copyright (c) 2019 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include "config.h"
#include "libfawk.h"
#include "parser.h"
#include "fpas.y.tab.h"
#include "lexer_inlines.h"

FAWK_API int fawk_lex_fpas(fpas_STYPE *lval, fawk_ctx_t *ctx)
{
	int chr, nchr;
	char tmp[128];

	restart:;

	if (ctx->parser.in_eof) goto handle_eof;
	lex_textblk_exec(PT_STRING, goto restart);

	/* start of a token: read the next non-space character */
	do { chr = getch(ctx); } while((chr == ' ') || (chr == '\t') || (chr == '\r') || (chr == '\n'));

	ctx->parser.used = 0;

	if (isalpha(chr) || (chr == '_')) { /* read an id or keyword */
		append(chr, return -1);
		for(;;) {
			chr = getch(ctx);
			if (!(isalpha(chr)) && !(isdigit(chr)) && (chr != '_') && (chr != '$')) {
				ungetch(ctx, chr);
				break;
			}
			append(chr, return -1);
			if (chr == '$') { append('\0', return -1); goto got_id; }
		}
		append('\0', return -1); /* append a virtual \0 so it's easier to strcmp */
		if      (strcmp(ctx->parser.buff, "if") == 0)      return PT_IF;
		else if (strcmp(ctx->parser.buff, "then") == 0)    return PT_THEN;
		else if (strcmp(ctx->parser.buff, "else") == 0)    return PT_ELSE;
		else if (strcmp(ctx->parser.buff, "exit") == 0)    return PT_EXIT;
		else if (strcmp(ctx->parser.buff, "for") == 0)     return PT_FOR;
		else if (strcmp(ctx->parser.buff, "down") == 0)    return PT_DOWN;
		else if (strcmp(ctx->parser.buff, "to") == 0)      return PT_TO;
		else if (strcmp(ctx->parser.buff, "while") == 0)   return PT_WHILE;
		else if (strcmp(ctx->parser.buff, "do") == 0)      return PT_DO;
		else if (strcmp(ctx->parser.buff, "repeat") == 0)  return PT_REPEAT;
		else if (strcmp(ctx->parser.buff, "until") == 0)   return PT_UNTIL;
		else if (strcmp(ctx->parser.buff, "function") == 0) return PT_FUNCTION;
		else if (strcmp(ctx->parser.buff, "begin") == 0)   return PT_BEGIN;
		else if (strcmp(ctx->parser.buff, "end") == 0)     return PT_END;
		else if (strcmp(ctx->parser.buff, "and") == 0)     return PT_AND;
		else if (strcmp(ctx->parser.buff, "or") == 0)      return PT_OR;
		else if (strcmp(ctx->parser.buff, "in") == 0)      return PT_IN;
		else if (strcmp(ctx->parser.buff, "not") == 0)     return PT_NOT;
		else if ((ctx->parser.curr_func != NULL) && (strcmp(ctx->parser.buff, ctx->parser.curr_func) == 0)) return PT_CURRFUNC;
		else if (strcmp(ctx->parser.buff, "include") == 0) { lex_include(goto restart); }
		else {
			got_id:; lval->str = fawk_strdup(ctx, ctx->parser.buff);
			return (lval->str == NULL) ? -1 : PT_ID;
		}
	}
	else if (chr ==  '{') { /* comment */
		readtil(ctx, "}"); getch(ctx);
		goto restart;
	}
	else if (isdigit(chr)) { /* read a number */
		append(chr, return -1);
		if (chr == '0') {
			nchr = getch(ctx);
			if (nchr == 'x') {
				fawk_readup(ctx, "0123456789abcdefABCDEF");
				lval->num = strtol(ctx->parser.buff, NULL, 16);
				return PT_NUM;
			}
			ungetch(ctx, nchr);
		}
		return read_numeric(ctx, &lval->num, 0, PT_NUM);
	}
	else switch(chr) {
		case '@': case '?': case ':': case ';': case '(': case ')':
		case '\\': case ']': case ',':
		case '+': case '-': case '*': case '/': case '%': case '=':
			return chr;

		case '[': { lex_textblk_start(goto restart); }

		case '.':
			nchr = getch(ctx);
			if (!isdigit(nchr)) {
				ungetch(ctx, nchr);
				return '.';
			}
			else {
				append('.', return -1); append(nchr, return -1);
				return read_numeric(ctx, &lval->num, 1, PT_NUM);
			}
			break;

		case '<':
			nchr = getch(ctx); append(chr, return -1);
			switch(nchr) {
				case '=': return PT_LTEQ;
				case '>': return PT_NEQ;
				default: ungetch(ctx, nchr); return chr;
			}
			break;

		case_op_ch2('>', '=', PT_GTEQ);

		case '\'':
			if (read_strlit(ctx, '\'') == 0) return PT_NIL;
			lval->str = fawk_strdup(ctx, ctx->parser.buff);
			return (lval->str == NULL) ? -1 : PT_STRING;

		case EOF:   { lex_got_eof(goto restart); }

		default:
			sprintf(tmp, "Invalid character on input: '%c'\n", chr);
			libfawk_error(ctx, tmp);
			return -1;
	}
}
